from benchling_api_client.v2.beta.api.dna_oligos import bulk_upsert_dna_oligos, upsert_dna_oligo
from benchling_api_client.v2.beta.models.dna_oligos_bulk_upsert_request import DnaOligosBulkUpsertRequest
from benchling_api_client.v2.beta.models.oligo_upsert_request import OligoUpsertRequest

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AsyncTaskLink, DnaOligo
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaDnaOligoService(BaseService):
    """
    V2-Beta DNA Oligos.

    DNA Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
    entities, they support schemas, tags, and aliases.

    See https://benchling.com/api/v2-beta/reference#/DNA%20Oligos
    """

    @api_method
    def upsert(self, entity_registry_id: str, dna_oligo: OligoUpsertRequest) -> DnaOligo:
        """
        Create or modify a DNA Oligo.

        See https://benchling.com/api/v2-beta/reference#/DNA%20Oligos/upsertDNAOligo
        """
        response = upsert_dna_oligo.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=dna_oligo
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(self, body: DnaOligosBulkUpsertRequest) -> AsyncTaskLink:
        """
        Bulk create or update DNA Oligos.

        See https://benchling.com/api/v2-beta/reference#/DNA%20Oligos/bulkUpsertDnaOligos
        """
        response = bulk_upsert_dna_oligos.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



from typing import Any, Dict, Iterable, List, Optional

from benchling_api_client.v2.stable.api.mixtures import (
    archive_mixtures,
    bulk_create_mixtures,
    bulk_update_mixtures,
    create_mixture,
    get_mixture,
    list_mixtures,
    unarchive_mixtures,
    update_mixture,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    EntityArchiveReason,
    ListMixturesSort,
    Mixture,
    MixtureBulkUpdate,
    MixtureCreate,
    MixturesArchivalChange,
    MixturesArchive,
    MixturesBulkCreateRequest,
    MixturesBulkUpdateRequest,
    MixturesPaginatedList,
    MixturesUnarchive,
    MixtureUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class MixtureService(BaseService):
    """
    Mixtures.

    Mixtures are solutions comprised of multiple ingredients where the exact quantities of each
    ingredient are important to track. Each ingredient is uniquely identified by its component entity.

    See https://benchling.com/api/reference#/Mixtures
    """

    @api_method
    def get_by_id(self, mixture_id: str) -> Mixture:
        """
        Get a mixture.

        See https://benchling.com/api/reference#/Mixtures/getMixture
        """
        response = get_mixture.sync_detailed(client=self.client, mixture_id=mixture_id)
        return model_from_detailed(response)

    @api_method
    def _mixtures_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[Iterable[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[Iterable[str]] = None,
        sort: Optional[ListMixturesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        ingredient_component_entity_ids: Optional[Iterable[str]] = None,
        ingredient_component_entity_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> Response[MixturesPaginatedList]:
        response = list_mixtures.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            ingredient_component_entity_ids=none_as_unset(
                optional_array_query_param(ingredient_component_entity_ids)
            ),
            ingredient_component_entity_idsany_of=none_as_unset(
                optional_array_query_param(ingredient_component_entity_ids_any_of)
            ),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[Iterable[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[Iterable[str]] = None,
        sort: Optional[ListMixturesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        ingredient_component_entity_ids: Optional[Iterable[str]] = None,
        ingredient_component_entity_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> PageIterator[Mixture]:
        """
        List Mixtures.

        See https://benchling.com/api/reference#/Mixtures/listMixtures
        """
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        def api_call(next_token: NextToken) -> Response[MixturesPaginatedList]:
            return self._mixtures_page(
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                ingredient_component_entity_ids=ingredient_component_entity_ids,
                ingredient_component_entity_ids_any_of=ingredient_component_entity_ids_any_of,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                schema_fields=schema_fields,
                sort=sort,
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
            )

        def results_extractor(body: MixturesPaginatedList) -> Optional[List[Mixture]]:
            return body.mixtures

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, mixture: MixtureCreate) -> Mixture:
        """
        Create a mixture.

        See https://benchling.com/api/reference#/Mixtures/createMixture
        """
        response = create_mixture.sync_detailed(client=self.client, json_body=mixture)
        return model_from_detailed(response)

    @api_method
    def update(self, mixture_id: str, mixture: MixtureUpdate) -> Mixture:
        """
        Update a mixture.

        See https://benchling.com/api/reference#/Mixtures/updateMixture
        """
        response = update_mixture.sync_detailed(client=self.client, mixture_id=mixture_id, json_body=mixture)
        return model_from_detailed(response)

    @api_method
    def archive(self, mixture_ids: Iterable[str], reason: EntityArchiveReason) -> MixturesArchivalChange:
        """
        Archive mixtures.

        See https://benchling.com/api/reference#/Mixtures/archiveMixtures
        """
        archive_request = MixturesArchive(reason=reason, mixture_ids=list(mixture_ids))
        response = archive_mixtures.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, mixture_ids: Iterable[str]) -> MixturesArchivalChange:
        """
        Unarchive mixtures.

        See https://benchling.com/api/reference#/Mixtures/unarchiveMixtures
        """
        unarchive_request = MixturesUnarchive(mixture_ids=list(mixture_ids))
        response = unarchive_mixtures.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_create(self, mixtures: Iterable[MixtureCreate]) -> AsyncTaskLink:
        """
        Bulk create mixtures.

        See https://benchling.com/api/reference#/Mixtures/bulkCreateMixtures
        """
        body = MixturesBulkCreateRequest(list(mixtures))
        response = bulk_create_mixtures.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, mixtures: Iterable[MixtureBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update mixtures.

        See https://benchling.com/api/reference#/Mixtures/bulkUpdateMixtures
        """
        body = MixturesBulkUpdateRequest(list(mixtures))
        response = bulk_update_mixtures.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



from datetime import date, datetime
from typing import List, Optional, Type, Union

from benchling_api_client.v2.alpha.models.base_manifest_config import BaseManifestConfig
from benchling_api_client.v2.alpha.models.dropdown_dependency import DropdownDependency
from benchling_api_client.v2.alpha.models.entity_schema_dependency import EntitySchemaDependency
from benchling_api_client.v2.alpha.models.field_definitions_manifest import FieldDefinitionsManifest
from benchling_api_client.v2.alpha.models.schema_dependency import SchemaDependency
from benchling_api_client.v2.alpha.models.schema_dependency_subtypes import SchemaDependencySubtypes
from benchling_api_client.v2.alpha.models.schema_dependency_types import SchemaDependencyTypes
from benchling_api_client.v2.alpha.models.workflow_task_schema_dependency import WorkflowTaskSchemaDependency
from benchling_api_client.v2.alpha.models.workflow_task_schema_dependency_output import (
    WorkflowTaskSchemaDependencyOutput,
)
from benchling_api_client.v2.beta.models.scalar_config import ScalarConfig
from benchling_api_client.v2.beta.models.scalar_config_types import ScalarConfigTypes
from benchling_api_client.v2.stable.extensions import NotPresentError

from benchling_sdk.apps.config.scalars import JsonType, ScalarModelType
from benchling_sdk.models import (
    AaSequence,
    AssayResult,
    AssayRun,
    Box,
    Container,
    CustomEntity,
    DnaOligo,
    DnaSequence,
    Entry,
    Location,
    Mixture,
    Molecule,
    Plate,
    Request,
    RnaOligo,
    RnaSequence,
    WorkflowTask,
)

_MODEL_TYPES_FROM_SCHEMA_TYPE = {
    SchemaDependencyTypes.CONTAINER_SCHEMA: Container,
    SchemaDependencyTypes.PLATE_SCHEMA: Plate,
    SchemaDependencyTypes.BOX_SCHEMA: Box,
    SchemaDependencyTypes.LOCATION_SCHEMA: Location,
    SchemaDependencyTypes.ENTRY_SCHEMA: Entry,
    SchemaDependencyTypes.REQUEST_SCHEMA: Request,
    SchemaDependencyTypes.RESULT_SCHEMA: AssayResult,
    SchemaDependencyTypes.RUN_SCHEMA: AssayRun,
    SchemaDependencyTypes.WORKFLOW_TASK_SCHEMA: WorkflowTask,
}


_SCALAR_TYPES_FROM_CONFIG = {
    ScalarConfigTypes.BOOLEAN: bool,
    ScalarConfigTypes.DATE: date,
    ScalarConfigTypes.DATETIME: datetime,
    ScalarConfigTypes.FLOAT: float,
    ScalarConfigTypes.INTEGER: int,
    ScalarConfigTypes.JSON: JsonType,
    ScalarConfigTypes.TEXT: str,
}

ModelType = Union[AssayResult, AssayRun, Box, Container, Entry, Location, Plate, Request]

_INSTANCE_FROM_SCHEMA_SUBTYPE = {
    SchemaDependencySubtypes.AA_SEQUENCE: AaSequence,
    SchemaDependencySubtypes.CUSTOM_ENTITY: CustomEntity,
    SchemaDependencySubtypes.DNA_SEQUENCE: DnaSequence,
    SchemaDependencySubtypes.DNA_OLIGO: DnaOligo,
    SchemaDependencySubtypes.MIXTURE: Mixture,
    SchemaDependencySubtypes.MOLECULE: Molecule,
    SchemaDependencySubtypes.RNA_OLIGO: RnaOligo,
    SchemaDependencySubtypes.RNA_SEQUENCE: RnaSequence,
}

EntitySubtype = Union[
    AaSequence, CustomEntity, DnaOligo, DnaSequence, Mixture, Molecule, RnaOligo, RnaSequence
]


class UnsupportedSubTypeError(Exception):
    """Error when an unsupported subtype is encountered."""

    pass


def model_type_from_dependency(
    dependency: Union[EntitySchemaDependency, SchemaDependency]
) -> Type[Union[ModelType, EntitySubtype]]:
    """Translate a schema dependency to its model class."""
    if isinstance(dependency, EntitySchemaDependency):
        model_type = _subtype_instance_from_dependency(dependency)
    else:
        model_type = _MODEL_TYPES_FROM_SCHEMA_TYPE[dependency.type]
    return model_type


def scalar_type_from_config(config: ScalarConfig) -> Union[object, Type[ScalarModelType]]:
    """Translate a scalar config to its Pyton type."""
    # We union with object to satisfy JsonType.
    return _SCALAR_TYPES_FROM_CONFIG.get(config.type, str)


def field_definitions_from_dependency(
    dependency: Union[
        EntitySchemaDependency,
        SchemaDependency,
        WorkflowTaskSchemaDependency,
        WorkflowTaskSchemaDependencyOutput,
    ]
) -> List[FieldDefinitionsManifest]:
    """Safely return a list of field definitions from a schema dependency or empty list."""
    try:
        if hasattr(dependency, "field_definitions"):
            return dependency.field_definitions
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def workflow_task_schema_output_from_dependency(
    dependency: WorkflowTaskSchemaDependency,
) -> Optional[WorkflowTaskSchemaDependencyOutput]:
    """Safely return a workflow task schema output from a workflow task schema or None."""
    try:
        if hasattr(dependency, "output"):
            return dependency.output
    # We can't seem to handle this programmatically by checking isinstance() or output truthiness
    except NotPresentError:
        pass
    return None


def options_from_dependency(dependency: DropdownDependency) -> List[BaseManifestConfig]:
    """Safely return a list of options from a dropdown dependency or empty list."""
    try:
        if hasattr(dependency, "options"):
            return dependency.options
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def _subtype_instance_from_dependency(dependency: EntitySchemaDependency) -> Type[EntitySubtype]:
    if dependency.subtype in _INSTANCE_FROM_SCHEMA_SUBTYPE:
        return _INSTANCE_FROM_SCHEMA_SUBTYPE[dependency.subtype]
    # This would mean the spec has a new subtype, the manifest installed in Benchling has declared it,
    # the user has linked it in Benchling, but the app code receiving it was never updated.
    # App developers should support it in deployed app code before republishing the manifest.
    raise UnsupportedSubTypeError(
        f"An unsupported subtype, {dependency.subtype.value} was received. "
        f"The version of the SDK may need to be upgraded to support this."
    )



from abc import ABC

from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.retry_helpers import RetryStrategy


class BaseService(ABC):
    """Abstract class for Benchling functional namespaces."""

    _client: Client
    _retry_strategy: RetryStrategy

    def __init__(self, client: Client, retry_strategy: RetryStrategy = RetryStrategy()):
        """
        Initialize a service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        self._client = client
        self._retry_strategy = retry_strategy

    @property
    def client(self) -> Client:
        """Provide access to the underlying generated Benchling Client."""
        return self._client

    @property
    def retry_strategy(self) -> RetryStrategy:
        """Provide access to the underlying user-specified RetryStrategy."""
        return self._retry_strategy



from typing import Iterable, List, Optional

from benchling_api_client.v2.alpha.api.apps import (
    archive_app_configuration_items,
    archive_canvases,
    bulk_create_app_configuration_items,
    bulk_update_app_configuration_items,
    create_app_configuration_item,
    create_canvas,
    get_app_configuration_item_by_id,
    get_benchling_app_manifest,
    get_canvas,
    list_app_configuration_items,
    put_benchling_app_manifest,
    unarchive_app_configuration_items,
    unarchive_canvases,
    update_app_configuration_item,
    update_canvas,
)
from benchling_api_client.v2.alpha.models.app_config_item import AppConfigItem
from benchling_api_client.v2.alpha.models.app_config_item_bulk_update import AppConfigItemBulkUpdate
from benchling_api_client.v2.alpha.models.app_config_item_create import AppConfigItemCreate
from benchling_api_client.v2.alpha.models.app_config_item_update import AppConfigItemUpdate
from benchling_api_client.v2.alpha.models.app_config_items_archival_change import AppConfigItemsArchivalChange
from benchling_api_client.v2.alpha.models.app_config_items_archive import AppConfigItemsArchive
from benchling_api_client.v2.alpha.models.app_config_items_archive_reason import AppConfigItemsArchiveReason
from benchling_api_client.v2.alpha.models.app_config_items_bulk_create_request import (
    AppConfigItemsBulkCreateRequest,
)
from benchling_api_client.v2.alpha.models.app_config_items_bulk_update_request import (
    AppConfigItemsBulkUpdateRequest,
)
from benchling_api_client.v2.alpha.models.app_config_items_unarchive import AppConfigItemsUnarchive
from benchling_api_client.v2.alpha.models.app_configuration_paginated_list import (
    AppConfigurationPaginatedList,
)
from benchling_api_client.v2.alpha.models.benchling_app_manifest import BenchlingAppManifest
from benchling_api_client.v2.alpha.models.canvas import Canvas
from benchling_api_client.v2.alpha.models.canvas_create import CanvasCreate
from benchling_api_client.v2.alpha.models.canvas_update import CanvasUpdate
from benchling_api_client.v2.alpha.models.canvases_archival_change import CanvasesArchivalChange
from benchling_api_client.v2.alpha.models.canvases_archive import CanvasesArchive
from benchling_api_client.v2.alpha.models.canvases_archive_reason import CanvasesArchiveReason
from benchling_api_client.v2.alpha.models.canvases_unarchive import CanvasesUnarchive
from benchling_api_client.v2.alpha.models.list_app_configuration_items_sort import (
    ListAppConfigurationItemsSort,
)
from benchling_api_client.v2.stable.types import Response

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import AsyncTaskLink
from benchling_sdk.services.v2.base_service import BaseService


class V2AlphaAppService(BaseService):
    """
    V2-Alpha Apps.

    Create and manage Apps on your tenant.

    https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps
    """

    @api_method
    def _list_app_configuration_items_page(
        self,
        *,
        app_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        modified_at: Optional[str] = None,
        sort: Optional[ListAppConfigurationItemsSort] = None,
        archive_reason: Optional[str] = None,
    ) -> Response[AppConfigurationPaginatedList]:
        return list_app_configuration_items.sync_detailed(  # type: ignore
            client=self.client,
            app_id=none_as_unset(app_id),
            ids=none_as_unset(optional_array_query_param(ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            modified_at=none_as_unset(modified_at),
            sort=none_as_unset(sort),
            archive_reason=none_as_unset(archive_reason),
        )

    def list_app_configuration_items(
        self,
        *,
        app_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = 50,
        modified_at: Optional[str] = None,
        sort: Optional[ListAppConfigurationItemsSort] = None,
        archive_reason: Optional[str] = None,
        # TODO BNCH-51788 Probably need TypeVar or Union[] instead of AppConfigItem
    ) -> PageIterator[AppConfigItem]:
        """
        Get app configuration items.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/listAppConfigurationItems
        """

        def api_call(next_token: NextToken) -> Response[AppConfigurationPaginatedList]:
            return self._list_app_configuration_items_page(
                app_id=app_id,
                ids=ids,
                page_size=page_size,
                modified_at=modified_at,
                sort=sort,
                archive_reason=archive_reason,
                next_token=next_token,
            )

        def results_extractor(
            body: AppConfigurationPaginatedList,
        ) -> Optional[List[AppConfigItem]]:
            return body.app_configuration_items

        return PageIterator(api_call, results_extractor)

    @api_method
    # TODO BNCH-51788 Probably need TypeVar or Union[] instead of AppConfigItem
    def get_app_configuration_item_by_id(self, item_id: str) -> AppConfigItem:
        """
        Get app configuration.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/getAppConfigurationItemById
        """
        response = get_app_configuration_item_by_id.sync_detailed(client=self.client, item_id=item_id)
        return model_from_detailed(response)

    @api_method
    # TODO BNCH-51788 Probably need TypeVar or Union[] instead of AppConfigItem and AppConfigItemCreate
    def create_app_configuration_item(self, configuration_item: AppConfigItemCreate) -> AppConfigItem:
        """
        Create app configuration item.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/createAppConfigurationItem
        """
        response = create_app_configuration_item.sync_detailed(
            client=self.client, json_body=configuration_item
        )
        return model_from_detailed(response)

    @api_method
    def update_app_configuration_item(
        # TODO BNCH-51788 Probably need TypeVar or Union[] instead of AppConfigItemUpdate
        self,
        item_id: str,
        configuration_item: AppConfigItemUpdate
        # TODO BNCH-51788 Probably need TypeVar or Union[] instead of AppConfigItem
    ) -> AppConfigItem:
        """
        Update app configuration item.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/updateBenchlingAppConfigurationItem
        """
        # TODO BNCH-51788 Should probably be renamed update_app_configuration_item for consistency
        response = update_app_configuration_item.sync_detailed(
            client=self.client, item_id=item_id, json_body=configuration_item
        )
        return model_from_detailed(response)

    @api_method
    def archive_app_configuration_items(
        self, item_ids: Iterable[str], reason: AppConfigItemsArchiveReason
    ) -> AppConfigItemsArchivalChange:
        """
        Archive app configuration items.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/archiveAppConfigurationItems
        """
        archive_request = AppConfigItemsArchive(reason=reason, item_ids=list(item_ids))
        response = archive_app_configuration_items.sync_detailed(
            client=self.client, json_body=archive_request
        )
        return model_from_detailed(response)

    @api_method
    def unarchive_app_configuration_items(self, item_ids: Iterable[str]) -> AppConfigItemsArchivalChange:
        """
        Unarchive app configuration items.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/unarchiveAppConfigurationItems
        """
        unarchive_request = AppConfigItemsUnarchive(item_ids=list(item_ids))
        response = unarchive_app_configuration_items.sync_detailed(
            client=self.client, json_body=unarchive_request
        )
        return model_from_detailed(response)

    @api_method
    def bulk_create_app_configuration_items(self, items: Iterable[AppConfigItemCreate]) -> AsyncTaskLink:
        """
        Bulk create app configuration items.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/bulkCreateAppConfigurationItems
        """
        body = AppConfigItemsBulkCreateRequest(list(items))
        response = bulk_create_app_configuration_items.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update_app_configuration_items(self, items: Iterable[AppConfigItemBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update app configuration items.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/bulkUpdateAppConfigurationItems
        """
        body = AppConfigItemsBulkUpdateRequest(list(items))
        response = bulk_update_app_configuration_items.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def get_manifest(self, app_id: str) -> BenchlingAppManifest:
        """
        Get app manifest.

        See https://benchling.com/api/v2-alpha/reference?stability=la/Apps/getBenchlingAppManifest
        """
        response = get_benchling_app_manifest.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)

    @api_method
    def update_manifest(self, app_id: str, manifest: BenchlingAppManifest) -> BenchlingAppManifest:
        """
        Update an app manifest.

        See https://benchling.com/api/v2-alpha/reference?stability=la#/Apps/putBenchlingAppManifest
        """
        response = put_benchling_app_manifest.sync_detailed(
            client=self.client, app_id=app_id, yaml_body=manifest
        )
        return model_from_detailed(response)

    @api_method
    def create_canvas(self, canvas: CanvasCreate) -> Canvas:
        """
        Create an App Canvas that a Benchling App can write to and read user interaction from.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/createCanvas
        """
        response = create_canvas.sync_detailed(
            client=self.client,
            json_body=canvas,
        )
        return model_from_detailed(response)

    @api_method
    def get_canvas(self, canvas_id: str) -> Canvas:
        """
        Get the current state of the App Canvas, including user input elements.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/getCanvas
        """
        response = get_canvas.sync_detailed(
            client=self.client,
            canvas_id=canvas_id,
        )
        return model_from_detailed(response)

    @api_method
    def update_canvas(self, canvas_id: str, canvas: CanvasUpdate) -> Canvas:
        """
        Update App Canvas.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/updateCanvas
        """
        response = update_canvas.sync_detailed(
            client=self.client,
            canvas_id=canvas_id,
            json_body=canvas,
        )
        return model_from_detailed(response)

    @api_method
    def archive_canvases(
        self, canvas_ids: Iterable[str], reason: CanvasesArchiveReason
    ) -> CanvasesArchivalChange:
        """
        Archive App Canvases.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/archiveCanvases
        """
        archive_request = CanvasesArchive(reason=reason, canvas_ids=list(canvas_ids))
        response = archive_canvases.sync_detailed(
            client=self.client,
            json_body=archive_request,
        )
        return model_from_detailed(response)

    @api_method
    def unarchive_canvases(self, canvas_ids: Iterable[str]) -> CanvasesArchivalChange:
        """
        Unarchive App Canvases.

        See https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps/unarchiveCanvases
        """
        unarchive_request = CanvasesUnarchive(canvas_ids=list(canvas_ids))
        response = unarchive_canvases.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date, datetime
import json
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from benchling_api_client.v2.alpha.models.boolean_app_config_item_type import BooleanAppConfigItemType
from benchling_api_client.v2.alpha.models.date_app_config_item_type import DateAppConfigItemType
from benchling_api_client.v2.alpha.models.datetime_app_config_item_type import DatetimeAppConfigItemType
from benchling_api_client.v2.alpha.models.float_app_config_item_type import FloatAppConfigItemType
from benchling_api_client.v2.alpha.models.generic_app_config_item_type import GenericAppConfigItemType
from benchling_api_client.v2.alpha.models.integer_app_config_item_type import IntegerAppConfigItemType
from benchling_api_client.v2.alpha.models.json_app_config_item_type import JsonAppConfigItemType
from benchling_api_client.v2.alpha.models.secure_text_app_config_item_type import SecureTextAppConfigItemType
from typing_extensions import Literal

JsonType = Union[Dict[str, Any], List[Any], str, int, float, bool]
ScalarType = TypeVar("ScalarType", bool, date, datetime, float, int, JsonType, str)
# JsonType support requires object to be unioned. Currently we do it inline.
ScalarModelType = Union[bool, date, datetime, float, int, str]
# Enum values cannot be used in literals, so copy the strings
ScalarConfigItemType = Literal[
    "boolean",  # BooleanAppConfigItemType.BOOLEAN,
    "date",  # DateAppConfigItemType.DATE,
    "datetime",  # DatetimeAppConfigItemType.DATETIME,
    "float",  # FloatAppConfigItemType.FLOAT,
    "text",  # GenericAppConfigItemType.TEXT,
    "integer",  # IntegerAppConfigItemType.INTEGER,
    "json",  # JsonAppConfigItemType.JSON,
    "secure_text",  # SecureTextAppConfigItemType.SECURE_TEXT,
]


class ScalarDefinition(ABC, Generic[ScalarType]):
    """
    Scalar definition.

    Map how ScalarConfigTypes values can be converted into corresponding Python types.
    """

    @classmethod
    def init(cls):
        """Init."""
        return cls()

    @abstractmethod
    def from_str(self, value: Optional[str]) -> Optional[ScalarType]:
        """
        From string.

        Given an optional string value of scalar configuration, produce an Optional instance of the
        specific ScalarType. For instance, converting str to int.

        Used when coercing Python types from string values in API responses.
        """
        pass


class BoolScalar(ScalarDefinition[bool]):
    """
    Bool Scalar.

    Turn a Boolean-like string value into bool. Any permutation of "true" - case insensitive - is interpreted
    as True. Any other non-empty string is False.
    """

    def from_str(self, value: Optional[str]) -> Optional[bool]:
        """Convert optional str to optional bool."""
        # Though the spec declares str, this is actually being sent in JSON as a real Boolean
        # So runtime check defensively
        if value is not None:
            if isinstance(value, bool):
                return value
            if value.lower() == "true":
                return True
            return False
        return None


class DateScalar(ScalarDefinition[date]):
    """
    Date Scalar.

    Turn an ISO formatted date like YYYY-MM-dd into a date.
    """

    def from_str(self, value: Optional[str]) -> Optional[date]:
        """Convert optional str to optional date."""
        return date.fromisoformat(value) if value is not None else None


class DateTimeScalar(ScalarDefinition[datetime]):
    """
    Date Time Scalar.

    Turn a date time string into datetime.
    """

    def from_str(self, value: Optional[str]) -> Optional[datetime]:
        """Convert optional str to optional datetime."""
        return datetime.strptime(value, self.expected_format()) if value is not None else None

    @staticmethod
    def expected_format() -> str:
        """Return the expected date mask for parsing string to datetime."""
        return "%Y-%m-%d %H:%M:%S %p"


class FloatScalar(ScalarDefinition[float]):
    """
    Float Scalar.

    Turn a string into float. Assumes the string, if not empty, is a valid floating point.
    """

    def from_str(self, value: Optional[str]) -> Optional[float]:
        """Convert optional str to optional float."""
        return float(value) if value is not None else None


class IntScalar(ScalarDefinition[int]):
    """
    Int Scalar.

    Turn a string into int. Assumes the string, if not empty, is a valid integer.
    """

    def from_str(self, value: Optional[str]) -> Optional[int]:
        """Convert optional str to optional int."""
        return int(value) if value is not None else None


class JsonScalar(ScalarDefinition[JsonType]):
    """
    Json Scalar.

    Turn a string into JSON. Assumes the string is a valid JSON string.
    """

    def from_str(self, value: Optional[str]) -> Optional[JsonType]:
        """Convert optional str to optional JsonType."""
        return json.loads(value) if value is not None else None


class TextScalar(ScalarDefinition[str]):
    """
    Text Scalar.

    Text is already a string, so no conversion is performed.
    """

    def from_str(self, value: Optional[str]) -> Optional[str]:
        """Convert optional str to optional str. Implemented to appease ScalarDefinition contract."""
        return value


class SecureTextScalar(TextScalar):
    """
    Secure Text Scalar.

    Text is already a string, so no conversion is performed.
    """

    pass


# Maps scalar types from the API into typed Python SDK scalar definitions
DEFAULT_SCALAR_DEFINITIONS: Dict[ScalarConfigItemType, ScalarDefinition] = {
    BooleanAppConfigItemType.BOOLEAN.value: BoolScalar.init(),
    DateAppConfigItemType.DATE.value: DateScalar.init(),
    DatetimeAppConfigItemType.DATETIME.value: DateTimeScalar.init(),
    FloatAppConfigItemType.FLOAT.value: FloatScalar.init(),
    IntegerAppConfigItemType.INTEGER.value: IntScalar.init(),
    JsonAppConfigItemType.JSON.value: JsonScalar.init(),
    SecureTextAppConfigItemType.SECURE_TEXT.value: SecureTextScalar.init(),
    GenericAppConfigItemType.TEXT.value: TextScalar.init(),
}



from __future__ import annotations

from datetime import date, datetime
import json
import random
import string
from typing import List, Optional, Union

from benchling_api_client.v2.alpha.models.app_config_item import AppConfigItem
from benchling_api_client.v2.alpha.models.base_manifest_config import BaseManifestConfig
from benchling_api_client.v2.alpha.models.benchling_app_manifest import BenchlingAppManifest
from benchling_api_client.v2.alpha.models.boolean_app_config_item import BooleanAppConfigItem
from benchling_api_client.v2.alpha.models.boolean_app_config_item_type import BooleanAppConfigItemType
from benchling_api_client.v2.alpha.models.date_app_config_item import DateAppConfigItem
from benchling_api_client.v2.alpha.models.date_app_config_item_type import DateAppConfigItemType
from benchling_api_client.v2.alpha.models.datetime_app_config_item import DatetimeAppConfigItem
from benchling_api_client.v2.alpha.models.datetime_app_config_item_type import DatetimeAppConfigItemType
from benchling_api_client.v2.alpha.models.dropdown_dependency import DropdownDependency
from benchling_api_client.v2.alpha.models.entity_schema_dependency import EntitySchemaDependency
from benchling_api_client.v2.alpha.models.field_app_config_item import FieldAppConfigItem
from benchling_api_client.v2.alpha.models.field_app_config_item_type import FieldAppConfigItemType
from benchling_api_client.v2.alpha.models.float_app_config_item import FloatAppConfigItem
from benchling_api_client.v2.alpha.models.float_app_config_item_type import FloatAppConfigItemType
from benchling_api_client.v2.alpha.models.generic_app_config_item import GenericAppConfigItem
from benchling_api_client.v2.alpha.models.generic_app_config_item_type import GenericAppConfigItemType
from benchling_api_client.v2.alpha.models.integer_app_config_item import IntegerAppConfigItem
from benchling_api_client.v2.alpha.models.integer_app_config_item_type import IntegerAppConfigItemType
from benchling_api_client.v2.alpha.models.json_app_config_item import JsonAppConfigItem
from benchling_api_client.v2.alpha.models.json_app_config_item_type import JsonAppConfigItemType
from benchling_api_client.v2.alpha.models.manifest_scalar_config import ManifestScalarConfig
from benchling_api_client.v2.alpha.models.resource_dependency import ResourceDependency
from benchling_api_client.v2.alpha.models.schema_dependency import SchemaDependency
from benchling_api_client.v2.alpha.models.secure_text_app_config_item import SecureTextAppConfigItem
from benchling_api_client.v2.alpha.models.secure_text_app_config_item_type import SecureTextAppConfigItemType
from benchling_api_client.v2.alpha.models.workflow_task_schema_dependency import WorkflowTaskSchemaDependency
from benchling_api_client.v2.beta.models.scalar_config_types import ScalarConfigTypes
from benchling_api_client.v2.stable.extensions import UnknownType

from benchling_sdk.apps.config.scalars import DateTimeScalar
from benchling_sdk.apps.helpers.config_helpers import (
    field_definitions_from_dependency,
    options_from_dependency,
    workflow_task_schema_output_from_dependency,
)

ManifestDependencies = Union[
    DropdownDependency,
    EntitySchemaDependency,
    ManifestScalarConfig,
    ResourceDependency,
    SchemaDependency,
    WorkflowTaskSchemaDependency,
    UnknownType,
]


def mock_app_config_items_from_manifest(manifest: BenchlingAppManifest) -> List[AppConfigItem]:
    """
    Mock Benchling App Config Items.

    This method accepts an app manifest model and creates mocked values for app the app config.

    The concrete mocked out values, such as API Ids and schema names are nonsensical and random,
    but are valid.

    Code should avoid relying on specific values or conventions (such as API prefixes). If
    specific dependency values need to be tested in isolation, the caller can selectively
    override the randomized values with with_dependency().
    """
    root_config_items = [_mock_dependency(dependency) for dependency in manifest.configuration]
    return [config_item for sub_config_items in root_config_items for config_item in sub_config_items]


def replace_mocked_config_item_by_path(
    original_config: List[AppConfigItem], replacement_path: List[str], replacement_item: AppConfigItem
) -> List[AppConfigItem]:
    """Return an updated list of app config items with a specific config item replaced with the provided mock."""
    replaced_config = [config for config in original_config if config.path != replacement_path]
    replaced_config.append(replacement_item)
    return replaced_config


def _mock_dependency(
    dependency: ManifestDependencies,
) -> List[AppConfigItem]:
    """Mock a dependency from its manifest definition."""
    if isinstance(dependency, DropdownDependency):
        config_item = GenericAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=GenericAppConfigItemType.DROPDOWN,
            value=_random_string("val_"),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency)
            for subdependency in options_from_dependency(dependency)
        ]
        return [config_item] + sub_items
    elif isinstance(dependency, EntitySchemaDependency):
        config_item = GenericAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=GenericAppConfigItemType(dependency.subtype),
            value=_random_string("val_"),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency)
            for subdependency in field_definitions_from_dependency(dependency)
        ]
        return [config_item] + sub_items
    elif isinstance(dependency, SchemaDependency):
        config_item = GenericAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=GenericAppConfigItemType(dependency.type),
            value=_random_string("val_"),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency)
            for subdependency in field_definitions_from_dependency(dependency)
        ]
        return [config_item] + sub_items
    elif isinstance(dependency, WorkflowTaskSchemaDependency):
        config_item = GenericAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=GenericAppConfigItemType.WORKFLOW_TASK_SCHEMA,
            value=_random_string("val_"),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency)
            for subdependency in field_definitions_from_dependency(dependency)
        ]
        workflow_task_output = workflow_task_schema_output_from_dependency(dependency)
        if workflow_task_output:
            output_fields = field_definitions_from_dependency(workflow_task_output)
            output_items = [
                _mock_workflow_output_subdependency(subdependency, dependency)
                for subdependency in output_fields
            ]
            sub_items += output_items
        return [config_item] + sub_items
    elif isinstance(dependency, ManifestScalarConfig):
        return [_mock_scalar_dependency(dependency)]
    elif isinstance(dependency, UnknownType):
        return UnknownType(value="Unknown")
    else:
        return [
            GenericAppConfigItem(
                id=_random_string("aci_"),
                path=[dependency.name],
                type=GenericAppConfigItemType(dependency.type),
                value=_random_string("val_"),
            )
        ]


def _mock_scalar_dependency(dependency: ManifestScalarConfig) -> AppConfigItem:
    if dependency.type == ScalarConfigTypes.BOOLEAN:
        return BooleanAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=BooleanAppConfigItemType.BOOLEAN,
            value=_mock_scalar_value(dependency.type),
        )
    elif dependency.type == ScalarConfigTypes.DATE:
        app_config = DateAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=DateAppConfigItemType.DATE,
            value=_mock_scalar_value(dependency.type),
        )
        return app_config
    elif dependency.type == ScalarConfigTypes.DATETIME:
        app_config = DatetimeAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=DatetimeAppConfigItemType.DATETIME,
            value=_mock_scalar_value(dependency.type),
        )
        return app_config
    elif dependency.type == ScalarConfigTypes.FLOAT:
        app_config = FloatAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=FloatAppConfigItemType.FLOAT,
            value=_mock_scalar_value(dependency.type),
        )
        return app_config
    elif dependency.type == ScalarConfigTypes.INTEGER:
        app_config = IntegerAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=IntegerAppConfigItemType.INTEGER,
            value=_mock_scalar_value(dependency.type),
        )
        return app_config
    elif dependency.type == ScalarConfigTypes.JSON:
        app_config = JsonAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=JsonAppConfigItemType.JSON,
            value=_mock_scalar_value(dependency.type),
        )
        return app_config
    elif dependency.type == ScalarConfigTypes.SECURE_TEXT:
        app_config = SecureTextAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=SecureTextAppConfigItemType.SECURE_TEXT,
            value=_mock_scalar_value(dependency.type),
        )
        return app_config
    else:
        return GenericAppConfigItem(
            id=_random_string("aci_"),
            path=[dependency.name],
            type=GenericAppConfigItemType.TEXT,
            value=_mock_scalar_value(dependency.type),
        )


def _mock_subdependency(subdependency: BaseManifestConfig, parent_config) -> AppConfigItem:
    if isinstance(parent_config, DropdownDependency):
        return GenericAppConfigItem(
            id=_random_string("aci_"),
            path=[parent_config.name, subdependency.name],
            type=GenericAppConfigItemType.DROPDOWN_OPTION,
            value=_random_string("opt_"),
        )
    elif isinstance(parent_config, (EntitySchemaDependency, SchemaDependency, WorkflowTaskSchemaDependency)):
        path = [parent_config.name, subdependency.name]
        app_config = FieldAppConfigItem(
            id=_random_string("aci_"),
            path=path,
            type=FieldAppConfigItemType.FIELD,
            value=_random_string("tsf_"),
        )
        return app_config


def _mock_workflow_output_subdependency(subdependency: BaseManifestConfig, parent_config) -> AppConfigItem:
    app_config = FieldAppConfigItem(
        id=_random_string("aci_"),
        path=[parent_config.name, "output", subdependency.name],
        type=FieldAppConfigItemType.FIELD,
        value=_random_string("tsf_"),
    )
    return app_config


def _mock_scalar_value(scalar_type: ScalarConfigTypes) -> Optional[str]:
    """Mock a scalar config value from its manifest definition."""
    if scalar_type == scalar_type.BOOLEAN:
        return "true"
    elif scalar_type == scalar_type.DATE:
        return date.today().strftime("%Y-%m-%d")
    elif scalar_type == scalar_type.DATETIME:
        return datetime.now().strftime(DateTimeScalar.expected_format())
    elif scalar_type == scalar_type.FLOAT:
        return str(random.random())
    elif scalar_type == scalar_type.INTEGER:
        return str(random.randint(-1000, 1000))
    elif scalar_type == scalar_type.JSON:
        return json.dumps(
            {_random_string(): [_random_string(), _random_string()], _random_string(): random.random()}
        )
    return _random_string()


def _random_string(prefix: str = "", random_length: int = 20) -> str:
    """Generate a randomized string up to a specified length with an optional prefix."""
    delimited_prefix = f"{prefix}-" if prefix else ""
    return f"{delimited_prefix}{''.join(random.choice(string.ascii_letters) for i in range(random_length))}"



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.assay_results import (
    abort_assay_results_transaction,
    archive_assay_results,
    bulk_get_assay_results,
    commit_assay_results_transaction,
    create_assay_results,
    create_assay_results_in_transaction,
    create_assay_results_transaction,
    get_assay_result,
    list_assay_results,
    unarchive_assay_results,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    array_query_param,
    none_as_unset,
    optional_array_query_param,
)
from benchling_sdk.helpers.transaction_manager import TransactionManager
from benchling_sdk.models import (
    AssayResult,
    AssayResultCreate,
    AssayResultIdsRequest,
    AssayResultIdsResponse,
    AssayResultsArchive,
    AssayResultsBulkCreateRequest,
    AssayResultsCreateResponse,
    AssayResultsPaginatedList,
    AssayResultTransactionCreateResponse,
    ListAssayResultsSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class AssayResultService(BaseService):
    """
    Assay Results.

    Results represent the output of assays that have been performed. You can customize the schemas of results to
    fit your needs. Results can link to runs, batches, and other types.

    See https://benchling.com/api/reference#/Assay%20Results
    """

    @api_method
    def get_by_id(self, assay_result_id: str) -> AssayResult:
        """
        Get a result.

        See https://benchling.com/api/reference#/Assay%20Results/getAssayResult
        """
        response = get_assay_result.sync_detailed(client=self.client, assay_result_id=assay_result_id)
        return model_from_detailed(response)

    @api_method
    def _assay_results_page(
        self,
        schema_id: str,
        min_created_time: Optional[int] = None,
        max_created_time: Optional[int] = None,
        entity_ids: Optional[Iterable[str]] = None,
        assay_run_ids: Optional[Iterable[str]] = None,
        next_token: NextToken = None,
        page_size: Optional[int] = None,
        ids: Optional[Iterable[str]] = None,
        storage_ids: Optional[Iterable[str]] = None,
        created_atgt: Optional[str] = None,
        created_atgte: Optional[str] = None,
        created_atlt: Optional[str] = None,
        created_atlte: Optional[str] = None,
        sort: Optional[ListAssayResultsSort] = None,
    ) -> Response[AssayResultsPaginatedList]:
        entity_ids_string = optional_array_query_param(entity_ids)
        assay_run_ids_string = optional_array_query_param(assay_run_ids)
        response = list_assay_results.sync_detailed(
            client=self.client,
            schema_id=schema_id,
            min_created_time=none_as_unset(min_created_time),
            max_created_time=none_as_unset(max_created_time),
            entity_ids=none_as_unset(entity_ids_string),
            assay_run_ids=none_as_unset(assay_run_ids_string),
            ids=none_as_unset(optional_array_query_param(ids)),
            storage_ids=none_as_unset(optional_array_query_param(storage_ids)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            created_atgt=none_as_unset(created_atgt),
            created_atgte=none_as_unset(created_atgte),
            created_atlt=none_as_unset(created_atlt),
            created_atlte=none_as_unset(created_atlte),
            sort=none_as_unset(sort),
        )
        raise_for_status(response)
        return response

    def list(
        self,
        schema_id: str,
        min_created_time: Optional[int] = None,
        max_created_time: Optional[int] = None,
        entity_ids: Optional[Iterable[str]] = None,
        assay_run_ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = None,
        ids: Optional[Iterable[str]] = None,
        storage_ids: Optional[Iterable[str]] = None,
        created_atgt: Optional[str] = None,
        created_atgte: Optional[str] = None,
        created_atlt: Optional[str] = None,
        created_atlte: Optional[str] = None,
        sort: Optional[ListAssayResultsSort] = None,
    ) -> PageIterator[AssayResult]:
        """
        List results.

        See https://benchling.com/api/reference#/Assay%20Results/listAssayResults
        """

        def api_call(next_token: NextToken) -> Response[AssayResultsPaginatedList]:
            return self._assay_results_page(
                schema_id=schema_id,
                min_created_time=min_created_time,
                max_created_time=max_created_time,
                entity_ids=entity_ids,
                assay_run_ids=assay_run_ids,
                ids=ids,
                storage_ids=storage_ids,
                created_atgt=created_atgt,
                created_atgte=created_atgte,
                created_atlt=created_atlt,
                created_atlte=created_atlte,
                next_token=next_token,
                page_size=page_size,
                sort=sort,
            )

        def results_extractor(body: AssayResultsPaginatedList) -> Optional[List[AssayResult]]:
            return body.assay_results

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(self, assay_result_ids: Iterable[str]) -> Optional[List[AssayResult]]:
        """
        Bulk get assay results.

        Up to 200 IDs can be specified at once.

        See https://benchling.com/api/reference#/Assay%20Results/bulkGetAssayResults
        """
        result_ids_string = array_query_param(assay_result_ids)
        response = bulk_get_assay_results.sync_detailed(
            client=self.client, assay_result_ids=result_ids_string
        )
        results_list = model_from_detailed(response)
        return results_list.assay_results

    @api_method
    def create(self, assay_results: Iterable[AssayResultCreate]) -> AssayResultsCreateResponse:
        """
        Create 1 or more results.

        See https://benchling.com/api/reference#/Assay%20Results/createAssayResults
        """
        create_results = AssayResultsBulkCreateRequest(assay_results=list(assay_results))
        response = create_assay_results.sync_detailed(client=self.client, json_body=create_results)
        return model_from_detailed(response)

    @api_method
    def archive(self, assay_result_ids: Iterable[str]) -> AssayResultIdsResponse:
        """
        Archive assay results.

        Only results that have not been added to a Notebook Entry can be Archived.
        Once results are attached to a notebook entry, they are tracked in the
        history of that notebook entry, and cannot be archived.

        See https://benchling.com/api/reference#/Assay%20Results/archiveAssayResults
        """
        archive_request = AssayResultsArchive(assay_result_ids=list(assay_result_ids))
        response = archive_assay_results.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, assay_result_ids: Iterable[str]) -> AssayResultIdsResponse:
        """
        Unarchive assay results.

        See https://benchling.com/api/reference#/Assay%20Results/unarchiveAssayResults
        """
        unarchive_request = AssayResultIdsRequest(assay_result_ids=list(assay_result_ids))
        response = unarchive_assay_results.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def create_transaction(self) -> AssayResultTransactionCreateResponse:
        """
        Create a transaction.

        See https://benchling.com/api/reference#/Assay%20Results/createAssayResultsTransaction
        """
        response = create_assay_results_transaction.sync_detailed(client=self.client)
        return model_from_detailed(response)

    @api_method
    def create_results_in_transaction(
        self, transaction_id: str, assay_results: Iterable[AssayResultCreate]
    ) -> AssayResultsCreateResponse:
        """
        Create results in a transaction.

        See https://benchling.com/api/reference#/Assay%20Results/createAssayResultsInTransaction
        """
        create_request = AssayResultsBulkCreateRequest(assay_results=list(assay_results))
        response = create_assay_results_in_transaction.sync_detailed(
            client=self.client, transaction_id=transaction_id, json_body=create_request
        )
        return model_from_detailed(response)

    @api_method
    def commit_transaction(self, transaction_id: str) -> AssayResultTransactionCreateResponse:
        """
        Commit results in an active transaction.

        Committing a transaction will cause all results that have been uploaded to be saved and visible to others.

        See https://benchling.com/api/reference#/Assay%20Results/commitAssayResultsTransaction
        """
        response = commit_assay_results_transaction.sync_detailed(
            client=self.client, transaction_id=transaction_id
        )
        return model_from_detailed(response)

    @api_method
    def abort_transaction(self, transaction_id: str) -> AssayResultTransactionCreateResponse:
        """
        Abort a transaction.

        Aborting a transaction will discard all uploaded results.

        See https://benchling.com/api/reference#/Assay%20Results/abortAssayResultsTransaction
        """
        response = abort_assay_results_transaction.sync_detailed(
            client=self.client, transaction_id=transaction_id
        )
        return model_from_detailed(response)

    def transaction_manager(self) -> TransactionManager:
        """
        Create a Python context manager for adding results within a transaction.

        When the context exits, the transaction manager will attempt to commit
        the transaction.

        If an unhandled error occurs within the context, the transaction manager
        will automatically attempt to abort the transaction.

        :return: A Python context manager for the transaction
        :rtype: TransactionManager
        """

        def create_trans() -> str:
            return self.create_transaction().id

        def abort_trans(transaction_id: str) -> str:
            return self.abort_transaction(transaction_id=transaction_id).id

        def commit_trans(transaction_id: str) -> str:
            return self.commit_transaction(transaction_id=transaction_id).id

        def append_rows(transaction_id: str, rows: Iterable[AssayResultCreate]) -> List[str]:
            return self.create_results_in_transaction(
                transaction_id=transaction_id, assay_results=rows
            ).assay_results

        return TransactionManager(
            create_transaction_call=create_trans,
            abort_transaction_call=abort_trans,
            commit_transaction_call=commit_trans,
            append_row_call=append_rows,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['benchling_sdk',
 'benchling_sdk.apps',
 'benchling_sdk.apps.config',
 'benchling_sdk.apps.helpers',
 'benchling_sdk.auth',
 'benchling_sdk.docs',
 'benchling_sdk.helpers',
 'benchling_sdk.models',
 'benchling_sdk.services',
 'benchling_sdk.services.v2',
 'benchling_sdk.services.v2.alpha',
 'benchling_sdk.services.v2.beta',
 'benchling_sdk.services.v2.stable']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'attrs>=20.1.0,<23',
 'backoff>=1.10.0,<2.0.0',
 'benchling-api-client==2.0.57',
 'dataclasses-json>=0.5.2,<0.6.0',
 'httpx>=0.23.0',
 'python-dateutil>=2.8.0,<3.0.0',
 'typing-extensions>=3.7.4,<5.0']

setup_kwargs = {
    'name': 'benchling-sdk',
    'version': '1.5.0a5',
    'description': 'SDK for interacting with the Benchling Platform.',
    'long_description': '# Benchling SDK\n\nA Python 3.7+ SDK for the [Benchling](https://www.benchling.com/) platform designed to provide typed, fluent\ninteractions with [Benchling APIs](https://docs.benchling.com/reference).\n\n## Installation\n\nInstall the dependency via [Poetry](https://python-poetry.org/) (if applicable):\n\n```bash\npoetry add benchling-sdk\n```\n \nOr [Pip](https://pypi.org/project/pip/):\n \n```bash\npip install benchling-sdk\n```\n\n## Documentation\n\nDocumentation for the SDK is kept up-to-date at [docs.benchling.com](https://docs.benchling.com), and you can get started with\nit using this guide:\n[https://docs.benchling.com/docs/getting-started-with-the-sdk](https://docs.benchling.com/docs/getting-started-with-the-sdk).\n\n## Support\n\nTo report issues with using the SDK, contact [support@benchling.com](mailto:support@benchling.com).\n',
    'author': 'Benchling Support',
    'author_email': 'support@benchling.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.base_service import BaseService
from benchling_sdk.services.v2.stable.aa_sequence_service import AaSequenceService
from benchling_sdk.services.v2.stable.api_service import ApiService
from benchling_sdk.services.v2.stable.app_service import AppService
from benchling_sdk.services.v2.stable.assay_result_service import AssayResultService
from benchling_sdk.services.v2.stable.assay_run_service import AssayRunService
from benchling_sdk.services.v2.stable.blob_service import BlobService
from benchling_sdk.services.v2.stable.box_service import BoxService
from benchling_sdk.services.v2.stable.container_service import ContainerService
from benchling_sdk.services.v2.stable.custom_entity_service import CustomEntityService
from benchling_sdk.services.v2.stable.dna_alignments_service import DnaAlignmentsService
from benchling_sdk.services.v2.stable.dna_oligo_service import DnaOligoService
from benchling_sdk.services.v2.stable.dna_sequence_service import DnaSequenceService
from benchling_sdk.services.v2.stable.dropdown_service import DropdownService
from benchling_sdk.services.v2.stable.entry_service import EntryService
from benchling_sdk.services.v2.stable.event_service import EventService
from benchling_sdk.services.v2.stable.export_service import ExportService
from benchling_sdk.services.v2.stable.feature_library_service import FeatureLibraryService
from benchling_sdk.services.v2.stable.folder_service import FolderService
from benchling_sdk.services.v2.stable.inventory_service import InventoryService
from benchling_sdk.services.v2.stable.lab_automation_service import LabAutomationService
from benchling_sdk.services.v2.stable.label_template_service import LabelTemplateService
from benchling_sdk.services.v2.stable.location_service import LocationService
from benchling_sdk.services.v2.stable.mixture_service import MixtureService
from benchling_sdk.services.v2.stable.molecule_service import MoleculeService
from benchling_sdk.services.v2.stable.nucleotide_alignments_service import NucleotideAlignmentsService
from benchling_sdk.services.v2.stable.oligo_service import OligoService
from benchling_sdk.services.v2.stable.organization_service import OrganizationService
from benchling_sdk.services.v2.stable.plate_service import PlateService
from benchling_sdk.services.v2.stable.printer_service import PrinterService
from benchling_sdk.services.v2.stable.project_service import ProjectService
from benchling_sdk.services.v2.stable.registry_service import RegistryService
from benchling_sdk.services.v2.stable.request_service import RequestService
from benchling_sdk.services.v2.stable.rna_oligo_service import RnaOligoService
from benchling_sdk.services.v2.stable.rna_sequence_service import RnaSequenceService
from benchling_sdk.services.v2.stable.schema_service import SchemaService
from benchling_sdk.services.v2.stable.task_service import TaskService
from benchling_sdk.services.v2.stable.team_service import TeamService
from benchling_sdk.services.v2.stable.user_service import UserService
from benchling_sdk.services.v2.stable.warehouse_service import WarehouseService
from benchling_sdk.services.v2.stable.workflow_output_service import WorkflowOutputService
from benchling_sdk.services.v2.stable.workflow_task_group_service import WorkflowTaskGroupService
from benchling_sdk.services.v2.stable.workflow_task_service import WorkflowTaskService


class V2StableService(BaseService):
    """
    V2 Stable.

    Namespace containing support for the V2 stable endpoints of the Benchling API.
    """

    _aa_sequence_service: AaSequenceService
    _api_service: ApiService
    _app_service: AppService
    _assay_result_service: AssayResultService
    _assay_run_service: AssayRunService
    _blob_service: BlobService
    _box_service: BoxService
    _container_service: ContainerService
    _custom_entity_service: CustomEntityService
    _dropdown_service: DropdownService
    _dna_alignments_service: DnaAlignmentsService
    _dna_oligo_service: DnaOligoService
    _dna_sequence_service: DnaSequenceService
    _entries_service: EntryService
    _event_service: EventService
    _export_service: ExportService
    _feature_library_service: FeatureLibraryService
    _folder_service: FolderService
    _inventory_service: InventoryService
    _lab_automation_service: LabAutomationService
    _label_templates_service: LabelTemplateService
    _location_service: LocationService
    _mixture_service: MixtureService
    _molecule_service: MoleculeService
    _nucleotide_alignments: NucleotideAlignmentsService
    _oligo_service: OligoService
    _organization_service: OrganizationService
    _plate_service: PlateService
    _printer_service: PrinterService
    _project_service: ProjectService
    _registry_service: RegistryService
    _request_service: RequestService
    _rna_oligo_service: RnaOligoService
    _rna_sequence_service: RnaSequenceService
    _schema_service: SchemaService
    _task_service: TaskService
    _team_service: TeamService
    _user_service: UserService
    _warehouse_service: WarehouseService
    _workflow_output_service: WorkflowOutputService
    _workflow_task_group_service: WorkflowTaskGroupService
    _workflow_task_service: WorkflowTaskService

    def __init__(self, client: Client, retry_strategy: RetryStrategy = RetryStrategy()):
        """
        Initialize a service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        self._aa_sequence_service = AaSequenceService(client, retry_strategy=retry_strategy)
        self._api_service = ApiService(client, retry_strategy=retry_strategy)
        self._app_service = AppService(client, retry_strategy=retry_strategy)
        self._assay_result_service = AssayResultService(client, retry_strategy=retry_strategy)
        self._assay_run_service = AssayRunService(client, retry_strategy=retry_strategy)
        self._blob_service = BlobService(client, retry_strategy=retry_strategy)
        self._box_service = BoxService(client, retry_strategy=retry_strategy)
        self._container_service = ContainerService(client, retry_strategy=retry_strategy)
        self._custom_entity_service = CustomEntityService(client, retry_strategy=retry_strategy)
        self._dna_alignments_service = DnaAlignmentsService(client, retry_strategy=retry_strategy)
        self._dna_oligo_service = DnaOligoService(client, retry_strategy=retry_strategy)
        self._dna_sequence_service = DnaSequenceService(client, retry_strategy=retry_strategy)
        self._dropdown_service = DropdownService(client, retry_strategy=retry_strategy)
        self._entries_service = EntryService(client, retry_strategy=retry_strategy)
        self._event_service = EventService(client, retry_strategy=retry_strategy)
        self._export_service = ExportService(client, retry_strategy=retry_strategy)
        self._feature_library_service = FeatureLibraryService(client, retry_strategy=retry_strategy)
        self._folder_service = FolderService(client, retry_strategy=retry_strategy)
        self._inventory_service = InventoryService(client, retry_strategy=retry_strategy)
        self._lab_automation_service = LabAutomationService(client, retry_strategy=retry_strategy)
        self._label_templates_service = LabelTemplateService(client, retry_strategy=retry_strategy)
        self._location_service = LocationService(client, retry_strategy=retry_strategy)
        self._mixture_service = MixtureService(client, retry_strategy=retry_strategy)
        self._molecule_service = MoleculeService(client, retry_strategy=retry_strategy)
        self._nucleotide_alignments = NucleotideAlignmentsService(client, retry_strategy=retry_strategy)
        self._oligo_service = OligoService(client, retry_strategy=retry_strategy)
        self._organization_service = OrganizationService(client, retry_strategy=retry_strategy)
        self._plate_service = PlateService(client, retry_strategy=retry_strategy)
        self._printer_service = PrinterService(client, retry_strategy=retry_strategy)
        self._project_service = ProjectService(client, retry_strategy=retry_strategy)
        self._registry_service = RegistryService(client, retry_strategy=retry_strategy)
        self._request_service = RequestService(client, retry_strategy=retry_strategy)
        self._rna_oligo_service = RnaOligoService(client, retry_strategy=retry_strategy)
        self._rna_sequence_service = RnaSequenceService(client, retry_strategy=retry_strategy)
        self._schema_service = SchemaService(client, retry_strategy=retry_strategy)
        self._task_service = TaskService(client, retry_strategy=retry_strategy)
        self._team_service = TeamService(client, retry_strategy=retry_strategy)
        self._user_service = UserService(client, retry_strategy=retry_strategy)
        self._warehouse_service = WarehouseService(client, retry_strategy=retry_strategy)
        self._workflow_output_service = WorkflowOutputService(client, retry_strategy=retry_strategy)
        self._workflow_task_group_service = WorkflowTaskGroupService(client, retry_strategy=retry_strategy)
        self._workflow_task_service = WorkflowTaskService(client, retry_strategy=retry_strategy)

    @property
    def aa_sequences(self) -> AaSequenceService:
        """
        AA Sequences.

        AA Sequences are the working units of cells that make everything run (they help make structures, catalyze
        reactions and allow for signaling - a kind of internal cell communication). On Benchling, these are comprised
        of a string of amino acids and collections of other attributes, such as annotations.

        See https://benchling.com/api/reference#/AA%20Sequences
        """
        return self._aa_sequence_service

    @property
    def api(self) -> ApiService:
        """
        Make custom API calls with the underlying BenchlingApiClient.

        A common use case for this is making calls to API endpoints which may not yet be supported in the current SDK
        release. It's capable of making more "generic" calls utilizing our authorization scheme, as well as supporting
        some simple serialization and deserialization for custom models.
        """
        return self._api_service

    @property
    def apps(self) -> AppService:
        """
        Apps.

        Apps provide a framework for you to customize your teams’ experiences on
        Benchling with custom applications.

        See https://benchling.com/api/reference#/Apps
        and https://docs.benchling.com/docs/getting-started-benchling-apps
        """
        return self._app_service

    @property
    def assay_results(self) -> AssayResultService:
        """
        Assay Results.

        Results represent the output of assays that have been performed. You can customize the schemas of results to
        fit your needs. Results can link to runs, batches, and other types.

        See https://benchling.com/api/reference#/Assay%20Results
        """
        return self._assay_result_service

    @property
    def assay_runs(self) -> AssayRunService:
        """
        Assay Runs.

        Runs capture the details / parameters of a run that was performed. Results are usually nested under a run.

        See https://benchling.com/api/reference#/Assay%20Runs
        """
        return self._assay_run_service

    @property
    def blobs(self) -> BlobService:
        """
        Blobs.

        Blobs are opaque files that can be linked to other items in Benchling, like assay runs or results. For example,
        you can upload a blob, then upload an assay result that links to that blob by ID. The blob will then appear as
        part of the assay result in the Benchling web UI.

        See https://benchling.com/api/reference#/Blobs
        """
        return self._blob_service

    @property
    def boxes(self) -> BoxService:
        """
        Boxes.

        Boxes are a structured storage type, consisting of a grid of positions that can each hold one container. Unlike
        locations, there are a maximum number of containers that a box can hold (one per position).

        Boxes are all associated with schemas, which define the type of the box (e.g. "10x10 Cryo Box") along with the
        fields that are tracked and the dimensions of the box.

        Like all storage, every Box has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Boxes
        """
        return self._box_service

    @property
    def containers(self) -> ContainerService:
        """
        Containers.

        Containers are the backbone of sample management in Benchling. They represent physical containers, such as
        tubes or wells, that hold quantities of biological samples (represented by the batches inside the container).
        The container itself tracks its total volume, and the concentration of every batch inside of it.

        Containers are all associated with schemas, which define the type of the container (e.g. "Tube") along with the
        fields that are tracked.

        Like all storage, every container has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Containers
        """
        return self._container_service

    @property
    def custom_entities(self) -> CustomEntityService:
        """
        Custom Entities.

        Benchling supports custom entities for biological entities that are neither sequences or proteins. Custom
        entities must have an entity schema set and can have both schema fields and custom fields.

        See https://benchling.com/api/reference#/Custom%20Entities
        """
        return self._custom_entity_service

    @property
    def dna_alignments(self) -> DnaAlignmentsService:
        """
        DNA Alignments.

        A DNA alignment is a Benchling object representing an alignment of multiple DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Alignments
        """
        return self._dna_alignments_service

    @property
    def dna_oligos(self) -> DnaOligoService:
        """
        DNA Oligos.

        DNA Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/reference#/DNA%20Oligos
        """
        return self._dna_oligo_service

    @property
    def dna_sequences(self) -> DnaSequenceService:
        """
        DNA Sequences.

        DNA sequences are the bread and butter of the Benchling Molecular Biology suite. On Benchling, these are
        comprised of a string of nucleotides and collections of other attributes, such as annotations and primers.

        See https://benchling.com/api/reference#/DNA%20Sequences
        """
        return self._dna_sequence_service

    @property
    def rna_sequences(self) -> RnaSequenceService:
        """
        RNA Sequences.

        Chains of linear, single stranded RNA that support most capabilities and attributes of DNA Sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences
        """
        return self._rna_sequence_service

    @property
    def dropdowns(self) -> DropdownService:
        """
        Dropdowns.

        Dropdowns are registry-wide enums. Use dropdowns to standardize on spelling and naming conventions, especially
        for important metadata like resistance markers.

        See https://benchling.com/api/reference#/Dropdowns
        """
        return self._dropdown_service

    @property
    def entries(self) -> EntryService:
        """
        Entries.

        Entries are rich text documents that allow you to capture all of your experimental data in one place.

        See https://benchling.com/api/reference#/Entries
        """
        return self._entries_service

    @property
    def events(self) -> EventService:
        """
        Events.

        The Events system allows external services to subscribe to events that are triggered in Benchling (e.g. plasmid
        registration, request submission, etc).

        See https://benchling.com/api/reference#/Events
        """
        return self._event_service

    @property
    def exports(self) -> ExportService:
        """
        Exports.

        Export a Notebook Entry.

        See https://benchling.com/api/reference#/Exports
        """
        return self._export_service

    @property
    def feature_libraries(self) -> FeatureLibraryService:
        """
        Feature Libraries.

        Feature Libraries are collections of shared canonical patterns that can be used to generate
        annotations on matching regions of DNA Sequences or AA Sequences.

        See https://benchling.com/api/reference#/Feature%20Libraries
        """
        return self._feature_library_service

    @property
    def folders(self) -> FolderService:
        """
        Folders.

        Manage folder objects.

        See https://benchling.com/api/reference#/Folders
        """
        return self._folder_service

    @property
    def inventory(self) -> InventoryService:
        """
        Inventory.

        Manage inventory wide objects.

        See https://benchling.com/api/reference#/Inventory
        """
        return self._inventory_service

    @property
    def lab_automation(self) -> LabAutomationService:
        """
        Lab Automation.

        Lab Automation endpoints support integration with lab instruments, and liquid handlers to create samples or
        results, and capture transfers between containers at scale.

        See https://benchling.com/api/reference#/Lab%20Automation
        """
        return self._lab_automation_service

    @property
    def label_templates(self) -> LabelTemplateService:
        """
        Label Templates.

        List label templates.

        See https://benchling.com/api/reference#/Label%20Templates
        """
        return self._label_templates_service

    @property
    def locations(self) -> LocationService:
        """
        Locations.

        Manage locations objects. Like all storage, every Location has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Locations
        """
        return self._location_service

    @property
    def mixtures(self) -> MixtureService:
        """
        Mixtures.

        Mixtures are solutions comprised of multiple ingredients where the exact quantities of each ingredient are
        important to track. Each ingredient is uniquely identified by its component entity.

        See https://benchling.com/api/reference#/Mixtures
        """
        return self._mixture_service

    @property
    def molecules(self) -> MoleculeService:
        """
        Molecules.

        Molecules are groups of atoms held together by bonds, representing entities smaller than DNA
        Sequences and AA Sequences. Just like other entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/reference#/Molecules
        """
        return self._molecule_service

    @property
    def nucleotide_alignments(self) -> NucleotideAlignmentsService:
        """
        Nucleotide Alignments.

        A Nucleotide Alignment is a Benchling object representing an alignment of multiple DNA and/or RNA sequences.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments
        """
        return self._nucleotide_alignments

    @property
    def oligos(self) -> OligoService:
        """
        Oligos.

        Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        Please migrate to the corresponding DNA Oligos endpoints so that we can support RNA Oligos.

        See https://benchling.com/api/reference#/Oligos
        """
        return self._oligo_service

    @property
    def organizations(self) -> OrganizationService:
        """
        Organizations.

        View organization objects.

        See https://benchling.com/api/reference#/Organizations
        """
        return self._organization_service

    @property
    def plates(self) -> PlateService:
        """
        Plates.

        Plates are a structured storage type, grids of wells that each function like containers. Plates come in two
        types: a traditional "fixed" type, where the wells cannot move, and a "matrix" type. A matrix plate has similar
        functionality to a box, where the containers inside can be moved around and removed altogether.

        Plates are all associated with schemas, which define the type of the plate (e.g. "96 Well Plate") along with
        the fields that are tracked, the dimensions of the plate, and whether or not the plate is a matrix plate or a
        traditional well plate.

        Like all storage, every Plate has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Plates
        """
        return self._plate_service

    @property
    def printers(self) -> PrinterService:
        """
        Printers.

        List printers.

        See https://benchling.com/api/reference#/Printers
        """
        return self._printer_service

    @property
    def projects(self) -> ProjectService:
        """
        Projects.

        Manage project objects.

        See https://benchling.com/api/reference#/Projects
        """
        return self._project_service

    @property
    def registry(self) -> RegistryService:
        """
        Registry.

        Manage registry objects.

        See https://benchling.com/api/reference#/Registry
        """
        return self._registry_service

    @property
    def requests(self) -> RequestService:
        """
        Requests.

        Requests allow scientists and teams to collaborate around experimental assays and workflows.

        See https://benchling.com/api/reference#/Requests
        """
        return self._request_service

    @property
    def rna_oligos(self) -> RnaOligoService:
        """
        RNA Oligos.

        RNA Oligos are short linear RNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/reference#/RNA%20Oligos
        """
        return self._rna_oligo_service

    @property
    def schemas(self) -> SchemaService:
        """
        Schemas.

        Schemas represent custom configuration of objects in Benchling. See https://docs.benchling.com/docs/schemas in
        our documentation on how Schemas impact our developers

        See https://benchling.com/api/reference#/Schemas
        """
        return self._schema_service

    @property
    def tasks(self) -> TaskService:
        """
        Tasks.

        Endpoints that perform expensive computations launch long-running tasks. These endpoints return the task ID (a
        UUID) in the response body.

        After launching a task, periodically invoke the Get a task endpoint with the task UUID (e.g., every 10
        seconds), until the status is no longer RUNNING.

        You can access a task for up to 30 minutes after its completion, after which its data will no longer be
        available.

        See https://benchling.com/api/reference#/Tasks
        """
        return self._task_service

    @property
    def teams(self) -> TeamService:
        """
        Teams.

        View team objects.

        See https://benchling.com/api/reference#/Teams
        """
        return self._team_service

    @property
    def users(self) -> UserService:
        """
        Benchling users.

        See https://benchling.com/api/reference#/Users
        """
        return self._user_service

    @property
    def warehouse(self) -> WarehouseService:
        """
        Warehouse.

        Manage warehouse credentials.

        See https://benchling.com/api/reference#/Warehouse
        """
        return self._warehouse_service

    @property
    def workflow_outputs(self) -> WorkflowOutputService:
        """
        Workflow Outputs.

        Workflow outputs are outputs of a workflow task.

        See https://benchling.com/api/reference#/Workflow%20Outputs
        """
        return self._workflow_output_service

    @property
    def workflow_task_groups(self) -> WorkflowTaskGroupService:
        """
        Workflow Tasks Groups.

        Workflow task groups are groups of workflow tasks of the same schema.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups
        """
        return self._workflow_task_group_service

    @property
    def workflow_tasks(self) -> WorkflowTaskService:
        """
        Workflow Tasks.

        Workflow tasks encapsulate a single unit of work.

        See https://benchling.com/api/reference#/Workflow%20Tasks
        """
        return self._workflow_task_service



from contextlib import AbstractContextManager
from typing import Any, Iterable, List

from typing_extensions import Protocol


class CreateTransactionCall(Protocol):
    """Provide an interface for creating a new transaction."""

    def __call__(self) -> str:
        """Create a transaction and return a string handle (i.e., id)."""
        ...


class CommitTransactionCall(Protocol):
    """Provide an interface for committing a transaction."""

    def __call__(self, transaction_id: str) -> str:
        """Commit an existing transaction in progress."""
        ...


class AbortTransactionCall(Protocol):
    """Provide an interface for aborting an uncommitted transaction."""

    def __call__(self, transaction_id: str) -> str:
        """Abort an existing transaction which has not been committed."""
        ...


class AppendRowsCall(Protocol):
    """Provide an interface for appending rows to an existing transaction."""

    def __call__(self, transaction_id: str, rows: Iterable[Any]) -> List[str]:
        """Append rows to an existing transaction and return the updated full list."""
        ...


class TransactionManager(AbstractContextManager):
    """
    Manage transactions in Benchling in a generic way across various implementations.

    Implements a Python context manager, managing state for a Benchling transaction. Upon
    exit, the transaction manager will automatically attempt to commit the transaction. If
    an error is encountered, it will instead abort the transaction and re-raise the error.
    """

    _transaction_id: str
    _create_transaction_call: CreateTransactionCall
    _commit_transaction_call: CommitTransactionCall
    _abort_transaction_call: AbortTransactionCall
    _append_row_call: AppendRowsCall

    def __init__(
        self,
        create_transaction_call: CreateTransactionCall,
        commit_transaction_call: CommitTransactionCall,
        abort_transaction_call: AbortTransactionCall,
        append_row_call: AppendRowsCall,
    ):
        """
        Initialize TransactionManager.

        :param create_transaction_call: A function for creating a transaction
        :param commit_transaction_call: A function for committing a transaction
        :param abort_transaction_call: A function for aborting a transaction
        :param append_row_call: A function for appending a row to an existing transaction
        """
        self._create_transaction_call = create_transaction_call
        self._commit_transaction_call = commit_transaction_call
        self._abort_transaction_call = abort_transaction_call
        self._append_row_call = append_row_call

    def __enter__(self) -> "TransactionManager":
        self._transaction_id = self._create_transaction_call()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
        if exc_type:
            self.abort()
            return False
        else:
            self.commit()
            return True

    def append(self, row: Any) -> List[str]:
        """Append a single row to an existing transaction and return the full list of pending rows."""
        return self.extend([row])

    def extend(self, rows: Iterable[Any]) -> List[str]:
        """Append multiple rows to an existing transaction and return the full list of pending rows."""
        return self._append_row_call(transaction_id=self._transaction_id, rows=rows)

    def commit(self) -> str:
        """Commit the transaction in progress."""
        return self._commit_transaction_call(self._transaction_id)

    def abort(self) -> str:
        """Abort the transaction in progress."""
        return self._abort_transaction_call(self._transaction_id)



from typing import Any, Dict, Iterable, List, Optional

from benchling_api_client.v2.stable.api.custom_entities import (
    archive_custom_entities,
    bulk_create_custom_entities,
    bulk_get_custom_entities,
    bulk_update_custom_entities,
    create_custom_entity,
    get_custom_entity,
    list_custom_entities,
    unarchive_custom_entities,
    update_custom_entity,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    CustomEntitiesArchivalChange,
    CustomEntitiesArchive,
    CustomEntitiesBulkCreateRequest,
    CustomEntitiesBulkUpdateRequest,
    CustomEntitiesPaginatedList,
    CustomEntitiesUnarchive,
    CustomEntity,
    CustomEntityBulkCreate,
    CustomEntityBulkUpdate,
    CustomEntityCreate,
    CustomEntityUpdate,
    EntityArchiveReason,
    ListCustomEntitiesSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class CustomEntityService(BaseService):
    """
    Custom Entities.

    Benchling supports custom entities for biological entities that are neither sequences or proteins. Custom
    entities must have an entity schema set and can have both schema fields and custom fields.

    See https://benchling.com/api/reference#/Custom%20Entities
    """

    @api_method
    def get_by_id(self, entity_id: str) -> CustomEntity:
        """
        Get a custom entity.

        See https://benchling.com/api/reference#/Custom%20Entities/getCustomEntity
        """
        response = get_custom_entity.sync_detailed(client=self.client, custom_entity_id=entity_id)
        return model_from_detailed(response)

    @api_method
    def _custom_entities_page(
        self,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListCustomEntitiesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
        author_idsany_of: Optional[Iterable[str]] = None,
        returning: Optional[List[str]] = None,
    ) -> Response[CustomEntitiesPaginatedList]:
        response = list_custom_entities.sync_detailed(
            client=self.client,
            schema_id=none_as_unset(schema_id),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        sort: Optional[ListCustomEntitiesSort] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
        returning: Optional[List[str]] = None,
    ) -> PageIterator[CustomEntity]:
        """
        List custom entities.

        See https://benchling.com/api/reference#/Custom%20Entities/listCustomEntities
        """
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        def api_call(next_token: NextToken) -> Response[CustomEntitiesPaginatedList]:
            return self._custom_entities_page(
                schema_id=schema_id,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                schema_fields=schema_fields,
                sort=sort,
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
                returning=returning,
            )

        def results_extractor(body: CustomEntitiesPaginatedList) -> Optional[List[CustomEntity]]:
            return body.custom_entities

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, entity: CustomEntityCreate) -> CustomEntity:
        """
        Create a custom entity.

        See https://benchling.com/api/reference#/Custom%20Entities/createCustomEntity
        """
        response = create_custom_entity.sync_detailed(client=self.client, json_body=entity)
        return model_from_detailed(response)

    @api_method
    def update(self, entity_id: str, entity: CustomEntityUpdate) -> CustomEntity:
        """
        Update a custom entity.

        See https://benchling.com/api/reference#/Custom%20Entities/updateCustomEntity
        """
        response = update_custom_entity.sync_detailed(
            client=self.client, custom_entity_id=entity_id, json_body=entity
        )
        return model_from_detailed(response)

    @api_method
    def archive(self, entity_ids: Iterable[str], reason: EntityArchiveReason) -> CustomEntitiesArchivalChange:
        """
        Archive custom entities.

        See https://benchling.com/api/reference#/Custom%20Entities/archiveCustomEntities
        """
        archive_request = CustomEntitiesArchive(reason=reason, custom_entity_ids=list(entity_ids))
        response = archive_custom_entities.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, entity_ids: Iterable[str]) -> CustomEntitiesArchivalChange:
        """
        Unarchive custom entities.

        See https://benchling.com/api/reference#/Custom%20Entities/unarchiveCustomEntities
        """
        unarchive_request = CustomEntitiesUnarchive(custom_entity_ids=list(entity_ids))
        response = unarchive_custom_entities.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_get(self, entity_ids: Iterable[str]) -> Optional[List[CustomEntity]]:
        """
        Bulk get custom entities.

        See https://benchling.com/api/reference#/Custom%20Entities/bulkGetCustomEntities
        """
        entity_id_string = ",".join(entity_ids)
        response = bulk_get_custom_entities.sync_detailed(
            client=self.client, custom_entity_ids=entity_id_string
        )
        custom_entities = model_from_detailed(response)
        return custom_entities.custom_entities

    @api_method
    def bulk_create(self, entities: Iterable[CustomEntityBulkCreate]) -> AsyncTaskLink:
        """
        Bulk create custom entities.

        See https://benchling.com/api/reference#/Custom%20Entities/bulkCreateCustomEntities
        """
        body = CustomEntitiesBulkCreateRequest(list(entities))
        response = bulk_create_custom_entities.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, entities: Iterable[CustomEntityBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update custom entities.

        See https://benchling.com/api/reference#/Custom%20Entities/bulkUpdateCustomEntities
        """
        body = CustomEntitiesBulkUpdateRequest(list(entities))
        response = bulk_update_custom_entities.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)
